from configure_vm_image.common.defaults import CONFIGURE_ARGUMENT
from configure_vm_image.cli.parsers.configure import configure_group


def configure_groups(parser):
    configure_group(parser)

    argument_groups = [CONFIGURE_ARGUMENT]
    return argument_groups



import os
from configure_vm_image.common.defaults import (
    CLOUD_INIT_DIR,
    TMP_DIR,
    RES_DIR,
)
from configure_vm_image.cli.parsers.configure import configure_group
from configure_vm_image.configure import configure_vm_image
from configure_vm_image.common.utils import expand_path


def configure_vm_image_cli(commands):
    parser = commands.add_parser(
        "configure-vm",
    )
    configure_group(parser)
    parser.set_defaults(func=corc_configure_vm_cli_exec)


def corc_configure_vm_cli_exec(args):
    image_path = args.get("image_path")
    image_format = args.get("image_format", None)
    config_user_data_path = args.get(
        "config_user_data_path", os.path.join(CLOUD_INIT_DIR, "user-data")
    )
    config_meta_data_path = args.get(
        "config_meta_data_path", os.path.join(CLOUD_INIT_DIR, "meta-data")
    )
    config_vendor_data_path = args.get(
        "config_vendor_data_path", os.path.join(CLOUD_INIT_DIR, "vendor-data")
    )
    config_network_config_path = args.get(
        "config_network_config_path", os.path.join(CLOUD_INIT_DIR, "network-config")
    )
    cloud_init_iso_output_path = args.get(
        "cloud_init_iso_output_path", os.path.join(CLOUD_INIT_DIR, "cidata.iso")
    )
    configure_vm_name = args.get("configure_vm_name", "configure-vm-image")
    configure_vm_log_path = args.get(
        "configure_vm_log_path", os.path.join(TMP_DIR, "configure-vm.log")
    )
    configure_vm_template_path = args.get(
        "configure_vm_template_path",
        os.path.join(RES_DIR, "configure-vm-template.xml.j2"),
    )
    configure_vm_template_values = args.get("configure_vm_template_values", {})
    reset_operations = args.get("reset_operations", "defaults,-ssh-userdir")
    verbose = args.get("verbose", False)

    return configure_vm_image(
        expand_path(image_path),
        image_format=image_format,
        user_data_path=expand_path(config_user_data_path),
        meta_data_path=expand_path(config_meta_data_path),
        vendor_data_path=expand_path(config_vendor_data_path),
        network_config_path=expand_path(config_network_config_path),
        cloud_init_iso_output_path=expand_path(cloud_init_iso_output_path),
        configure_vm_name=configure_vm_name,
        configure_vm_log_path=expand_path(configure_vm_log_path),
        configure_vm_template_path=expand_path(configure_vm_template_path),
        configure_vm_template_values=configure_vm_template_values,
        reset_operations=reset_operations,
        verbose=verbose,
    )



import os
from gen_vm_image.common.codes import SUCCESS
from gen_vm_image.image import generate_image
from gen_vm_image.utils.net import download_file
from configure_vm_image.utils.io import join, makedirs, remove, exists, load
from configure_vm_image.common.defaults import CPU_ARCHITECTURE


TEST_IMAGE_NAME = "test_image"
TEST_IMAGE_FORMAT = "qcow2"
TEST_IMAGE_TEMPLATE_CONFIG = "configure-vm-template.xml.j2"
TEST_IMAGE_VERSION = "9.5-20241118.0"

INPUT_IMAGE_URL = f"https://sid.erda.dk/share_redirect/B0AM7a5lpC/Rocky-9-GenericCloud-Base-{TEST_IMAGE_VERSION}.{CPU_ARCHITECTURE}.{TEST_IMAGE_FORMAT}"
INPUT_IMAGE_CHECKSUM_URL = f"https://sid.erda.dk/share_redirect/B0AM7a5lpC/Rocky-9-GenericCloud-Base-{TEST_IMAGE_VERSION}.{CPU_ARCHITECTURE}.{TEST_IMAGE_FORMAT}.chksum.txt"


def extract_checksum_from_image_file(path):
    """Extract checksum, expects the format 'checksum (of first xxxxx bytes) relative_serverside_path'"""
    content = load(path)
    if not content:
        return False
    split_content = content.split(" ")
    checksum = split_content[0]
    checksum_byte_size = split_content[3]
    return checksum, checksum_byte_size


class AsyncConfigureTestContext:
    def __init__(self):
        self.init_done = False

    async def setUp(self):
        if self.init_done:
            return

        # https://cloud.debian.org/images/cloud/bookworm/latest/input_path_image-genericcloud-amd64.qcow2
        # Download the test image for the path input test
        self.test_tmp_directory = os.path.realpath(join("tests", "tmp"))
        self.test_res_directory = os.path.realpath(join("tests", "res"))
        self.test_res_template_directory = os.path.realpath(
            join(self.test_res_directory, "templates")
        )
        if not exists(self.test_tmp_directory):
            assert makedirs(self.test_tmp_directory)

        self.image_config = {
            "name": TEST_IMAGE_NAME,
            # TODO, dynamically discover the size of the input image
            # This output size should not be less than what the original input disk image size
            "size": "10G",
            "input": INPUT_IMAGE_URL,
            "output_directory": self.test_tmp_directory,
            "output_format": TEST_IMAGE_FORMAT,
        }

        self.image = join(
            self.test_tmp_directory, "{}.{}".format(TEST_IMAGE_NAME, TEST_IMAGE_FORMAT)
        )
        if not exists(self.image):
            checksum_file_path = join(
                self.test_tmp_directory, f"{TEST_IMAGE_NAME}.checksum"
            )
            downloaded, response = await download_file(
                INPUT_IMAGE_CHECKSUM_URL, checksum_file_path
            )
            assert downloaded
            assert exists(checksum_file_path)
            checksum, checksum_byte_size = extract_checksum_from_image_file(
                checksum_file_path
            )
            # Test that the image can be configured
            success, msg = await generate_image(
                self.image_config["name"],
                self.image_config["size"],
                input=self.image_config["input"],
                input_checksum=checksum,
                input_checksum_type="sha256",
                input_checksum_read_bytes=int(checksum_byte_size),
                output_directory=self.image_config["output_directory"],
                output_format=self.image_config["output_format"],
            )
            assert success == SUCCESS
        assert exists(self.image)

        self.image_template_config = join(
            self.test_res_template_directory, TEST_IMAGE_TEMPLATE_CONFIG
        )
        assert exists(self.image_template_config)

    # Should be used by the non async function tearDownClass to ensure that
    # the following cleanup is done before the class is destroyed
    def tearDown(self):
        assert remove(self.test_tmp_directory, recursive=True)



import asyncio
import argparse
import inspect
import sys
import json
from configure_vm_image._version import __version__
from configure_vm_image.common.defaults import CONFIGURE_ARGUMENT
from configure_vm_image.common.utils import to_str, error_print
from configure_vm_image.common.codes import (
    SUCCESS,
    JSON_DUMP_ERROR,
    JSON_DUMP_ERROR_MSG,
)
from configure_vm_image.cli.helpers import (
    extract_arguments,
    strip_argument_group_prefix,
)

SCRIPT_NAME = __file__


def import_from_module(module_path, module_name, func_name):
    module = __import__(module_path, fromlist=[module_name])
    return getattr(module, func_name)


def add_cli_operations(
    parser,
    operation,
    module_cli_input_group_prefix="configure_vm_image.cli.input_groups",
    module_operation_prefix="configure_vm_image.cli.operations",
):
    operation_input_groups_func = import_from_module(
        "{}.{}".format(module_cli_input_group_prefix, operation),
        "{}".format(operation),
        "{}_groups".format(operation),
    )

    argument_groups = operation_input_groups_func(parser)

    parser.set_defaults(
        func=cli_exec,
        module_path="{}.{}".format(module_operation_prefix, operation),
        module_name="{}".format(operation),
        func_name="{}_operation".format(operation),
        argument_groups=argument_groups,
    )


def cli_exec(arguments):
    # Actions determines which function to execute
    module_path = arguments.pop("module_path")
    module_name = arguments.pop("module_name")
    func_name = arguments.pop("func_name")

    if "positional_arguments" in arguments:
        positional_arguments = arguments.pop("positional_arguments")
    else:
        positional_arguments = []

    if "argument_groups" in arguments:
        argument_groups = arguments.pop("argument_groups")
    else:
        argument_groups = []

    func = import_from_module(module_path, module_name, func_name)
    if not func:
        return False, {}

    action_kwargs, _ = extract_arguments(arguments, argument_groups)
    action_kwargs = strip_argument_group_prefix(action_kwargs, argument_groups)

    action_args = positional_arguments
    if inspect.iscoroutinefunction(func):
        return asyncio.run(func(*action_args, **action_kwargs))
    return func(*action_args, **action_kwargs)


def add_base_cli_operations(parser):
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help="Print the version of the program",
    )


def main(args):
    parser = argparse.ArgumentParser(
        prog=SCRIPT_NAME,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # Add the basic CLI functions
    add_base_cli_operations(parser)
    # Add the configure image CLI
    add_cli_operations(parser, "configure")

    parsed_args = parser.parse_args(args)
    # Convert to a dictionary
    arguments = vars(parsed_args)

    if "func" not in arguments:
        raise ValueError("Missing function to execute in prepared arguments")

    func = arguments.pop("func")
    return_code, result_dict = func(arguments)

    response = {}
    if return_code == SUCCESS:
        response["status"] = "success"
    else:
        response["status"] = "failed"
    if arguments.get("{}_verbose".format(CONFIGURE_ARGUMENT), False):
        response["outputs"] = result_dict.get("verbose_outputs", [])
    response["msg"] = result_dict.get("msg", "")
    response["return_code"] = return_code

    try:
        output = json.dumps(response, indent=4, sort_keys=True, default=to_str)
    except Exception as err:
        error_print(JSON_DUMP_ERROR_MSG.format(err))
        return JSON_DUMP_ERROR
    if return_code == SUCCESS:
        print(output)
    else:
        error_print(output)
    return return_code


def cli():
    return main(sys.argv[1:])


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import unittest
from configure_vm_image.common.codes import SUCCESS
from configure_vm_image.cli.configure_image import main


class TestCLIBase(unittest.TestCase):

    def test_cli_help(self):
        return_code = None
        try:
            return_code = main(["--help"])
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)

    def test_cli_version(self):
        return_code = None
        try:
            return_code = main(["--version"])
        except SystemExit as e:
            return_code = e.code
        self.assertEqual(return_code, SUCCESS)



import argparse


class PositionalArgumentsAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if not hasattr(namespace, "positional_arguments") or not getattr(
            namespace, "positional_arguments"
        ):
            setattr(namespace, "positional_arguments", [values])
        else:
            getattr(namespace, "positional_arguments").append(values)


class KeyValueAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        key_value_list = values.split(",")
        key_value_dict = {
            key_value.split("=")[0]: key_value.split("=")[1]
            for key_value in key_value_list
        }
        setattr(namespace, self.dest, key_value_dict)



import os
import time
from configure_vm_image.common.defaults import (
    CLOUD_INIT_DIR,
    TMP_DIR,
    RES_DIR,
    CONFIGURE_VM_VCPUS,
    CONFIGURE_VM_MEMORY,
    CONFIGURE_VM_MACHINE,
    CPU_ARCHITECTURE,
)
from configure_vm_image.common.codes import (
    SUCCESS,
    CONFIGURE_IMAGE_ERROR,
    CONFIGURE_IMAGE_ERROR_MSG,
    PATH_CREATE_ERROR,
    PATH_CREATE_ERROR_MSG,
    PATH_NOT_FOUND_ERROR,
    PATH_NOT_FOUND_ERROR_MSG,
    RESET_IMAGE_ERROR,
    RESET_IMAGE_ERROR_MSG,
)
from configure_vm_image.utils.job import run
from configure_vm_image.utils.io import exists, which, load, makedirs


def discover_create_iso_command():
    """Discovers the command to generate an iso on the system"""
    if os.uname().sysname == "Darwin":
        create_iso_command = "mkisofs"
    else:
        create_iso_command = "genisoimage"
    if not which(create_iso_command):
        raise FileNotFoundError(
            "Failed to find the {} command on the system. Please ensure that it is installed".format(
                create_iso_command
            )
        )
    return create_iso_command


def create_cloud_init_disk(
    output_path,
    user_data_path=None,
    meta_data_path=None,
    vendor_data_path=None,
    network_config_path=None,
):
    # Generated the configuration iso image
    # Notice that we label the iso cidata to ensure that cloud-init
    # recognizes the disk as a configuration disk
    create_iso_command = discover_create_iso_command()
    cloud_init_command = [
        create_iso_command,
        "-output",
        output_path,
        "-V",
        "cidata",
        "--joliet",
        "--rock",
    ]

    if user_data_path:
        cloud_init_command.append(user_data_path)
    if meta_data_path:
        cloud_init_command.append(meta_data_path)
    if vendor_data_path:
        cloud_init_command.append(vendor_data_path)
    if network_config_path:
        cloud_init_command.append(network_config_path)
    success, result = run(cloud_init_command)
    if not success:
        return PATH_CREATE_ERROR, PATH_CREATE_ERROR_MSG.format(
            output_path, result["error"]
        )
    return True, result["output"]


def virt_customize(image_path, commands_from_file):
    if not exists(image_path):
        return PATH_NOT_FOUND_ERROR, PATH_NOT_FOUND_ERROR_MSG.format(
            image_path, "could not find the image path to customize"
        )

    if not exists(commands_from_file):
        return PATH_NOT_FOUND_ERROR, PATH_NOT_FOUND_ERROR_MSG.format(
            commands_from_file,
            "could not find the commands file to customize the image with",
        )

    # Run the virt-customize command
    virt_customize_command = [
        "virt-customize",
        "-a",
        image_path,
        "--commands-from-file",
        commands_from_file,
    ]
    result = run(virt_customize_command)
    if result["returncode"] != "0":
        return CONFIGURE_IMAGE_ERROR, CONFIGURE_IMAGE_ERROR_MSG.format(
            image_path, result["error"]
        )

    return result, None


def generate_image_configuration(
    output_path,
    user_data_path=None,
    meta_data_path=None,
    vendor_data_path=None,
    network_config_path=None,
):
    return create_cloud_init_disk(
        output_path,
        user_data_path=user_data_path,
        meta_data_path=meta_data_path,
        vendor_data_path=vendor_data_path,
        network_config_path=network_config_path,
    )


def discover_vm_orchestrator():
    """Discovers the vm orchestrator command line tool on the system"""
    # TODO, accept the orchestrator as an argument to be dynamically discovered
    orchestrator = "libvirt-provider"
    if not which(orchestrator):
        raise FileNotFoundError(
            "Failed to find the {} command on the system. Please ensure that it is installed".format(
                orchestrator
            )
        )
    return orchestrator


async def configure_vm(name, image, template_path=None, **kwargs):
    """This launches a subprocess that configures the VM image on boot."""
    vm_orchestrator = discover_vm_orchestrator()
    # TODO discover the specific vm orchestrator argument structure
    create_command = [
        vm_orchestrator,
        "instance",
        "create",
        name,
        image,
    ]
    if template_path:
        create_command.extend(["--template-path", template_path])
        create_command.extend(["--extra-template-path-values"])
        create_command.append(
            ",".join([f"{key}={value}" for key, value in kwargs.items()])
        )

    create_success, create_result = run(create_command, output_format="json")
    if not create_success:
        return False, create_result["error"]

    if not isinstance(create_result, dict):
        return False, create_result

    if "error" in create_result and create_result["error"]:
        return False, create_result["error"]

    if "output" in create_result and not create_result["output"]:
        return False, create_result["output"]

    if not isinstance(create_result["output"], dict):
        return False, create_result["output"]

    if "instance" not in create_result["output"]:
        return False, create_result["output"]

    if not isinstance(create_result["output"]["instance"], dict):
        return False, create_result["output"]

    if "id" not in create_result["output"]["instance"]:
        return False, create_result["output"]["instance"]

    instance_id = create_result["output"]["instance"]["id"]
    start_command = [
        vm_orchestrator,
        "instance",
        "start",
        instance_id,
    ]
    start_success, start_result = run(start_command, output_format="json")
    if not start_success:
        return False, start_result["error"]
    return instance_id, start_result["output"]


async def configure_image(name, image, **kwargs):
    """Configures the image using the configuration path"""
    configure_result, configure_msg = await configure_vm(name, image, **kwargs)
    if not configure_result:
        return False, configure_msg
    return configure_result, configure_msg


def finished_configure(configure_vm_log_path):
    """Waits for the configuration process to finish"""
    # Wait for the configuration process to finish
    if not exists(configure_vm_log_path):
        return False

    first_marker = "Cloud-init v"
    second_marker = "finished at"

    finished = False
    while not finished:
        content = load(configure_vm_log_path, readlines=True)
        if content and isinstance(content, (list, set, tuple)):
            for line in content:
                if first_marker in line and second_marker in line:
                    finished = True
    return finished


def vm_action(action, name, *args, **kwargs):
    vm_orchestrator = discover_vm_orchestrator()
    command = [vm_orchestrator, "instance", action, name, *args]
    for key, value in kwargs.items():
        if key and value:
            command.extend([key, value])
        elif key and not value:
            command.append(key)
        elif value and not key:
            command.append(value)
    success, result = run(command, output_format="json")
    if not success:
        return False, result["error"]
    return True, result["output"]


def wait_for_vm_shutdown(name, attempts=30):
    """Waits for the VM to be shutdown"""
    attempt = 0
    while attempt < attempts:
        found, result = vm_action("show", name)
        if found:
            instance = result.get("instance", {})
            state = instance.get("state", "")
            if state == "shut off":
                return True, f"VM: {name} was successfully shutdown"
        else:
            return True, f"VM: {name} was is already removed"
        time.sleep(1)
        attempt += 1
    return False, f"Failed to wait for the shutdown of VM: {name}"


def wait_for_vm_removed(name, attempts=30):
    """Waits for the VM to be removed"""
    attempt = 0
    msg = ""
    while attempt < attempts:
        found, msg = vm_action("show", name)
        if not found:
            return True, f"VM: {name} was sucessfully removed"
        time.sleep(1)
        attempt += 1
    if not msg:
        msg = f"Failed to wait for the removal of VM: {name}"
    return False, msg


def reset_image(image, reset_operations=None, verbose=False):
    """Resets the image such that it is ready to be started
    in production"""
    # Ensure that the virt-sysprep doesn't try to use libvirt
    # but qemu instead
    # LIBGUESTFS_BACKEND=direct
    reset_command = ["virt-sysprep", "-a", image]
    if reset_operations:
        reset_command.extend(["--operations", reset_operations])
    if verbose:
        reset_command.append("--verbose")
    success, result = run(reset_command)
    if not success:
        return False, result["error"]
    return True, result["output"]


async def configure_vm_image(
    image_path,
    image_format=None,
    user_data_path=os.path.join(CLOUD_INIT_DIR, "user-data"),
    meta_data_path=os.path.join(CLOUD_INIT_DIR, "meta-data"),
    vendor_data_path=os.path.join(CLOUD_INIT_DIR, "vendor-data"),
    network_config_path=os.path.join(CLOUD_INIT_DIR, "network-config"),
    cloud_init_iso_output_path=os.path.join(CLOUD_INIT_DIR, "cidata.iso"),
    configure_vm_orchestrator="libvirt-provider",
    configure_vm_name="configure-vm-image",
    configure_vm_log_path=os.path.join(TMP_DIR, "configure-vm.log"),
    configure_vm_template_path=os.path.join(RES_DIR, "configure-vm-template.xml.j2"),
    configure_vm_template_values=None,
    reset_operations="defaults,-ssh-userdir",
    verbose=False,
):
    response = {}
    verbose_outputs = []

    if not configure_vm_template_values:
        configure_vm_template_values = {}

    # Ensure that the image to configure exists
    if not exists(image_path):
        response["msg"] = PATH_NOT_FOUND_ERROR_MSG.format(
            image_path, "could not find the image to configure"
        )
        response["verbose_outputs"] = verbose_outputs
        return PATH_NOT_FOUND_ERROR, response

    if not image_format:
        image_format = os.path.splitext(image_path)[1].replace(".", "")
        if verbose:
            verbose_outputs.append(
                "Automatically discovered image format: {} to configure the disk image".format(
                    image_format
                )
            )

    # Ensure that the required output directories exists
    cidata_iso_dir = os.path.dirname(cloud_init_iso_output_path)
    for d in [cidata_iso_dir]:
        if not exists(d):
            created = makedirs(d)
            if not created:
                response["msg"] = PATH_CREATE_ERROR_MSG.format(d)
                response["verbose_outputs"] = verbose_outputs
                return PATH_CREATE_ERROR, response

    if not exists(user_data_path):
        verbose_outputs.append(
            PATH_NOT_FOUND_ERROR_MSG.format(
                user_data_path,
                "could not find the user-data configuration file, continuing without it",
            )
        )
        user_data_path = None

    if not exists(meta_data_path):
        verbose_outputs.append(
            PATH_NOT_FOUND_ERROR_MSG.format(
                meta_data_path,
                "could not find the meta-data configuration file, continuing without it",
            )
        )
        meta_data_path = None

    if not exists(vendor_data_path):
        verbose_outputs.append(
            PATH_NOT_FOUND_ERROR_MSG.format(
                vendor_data_path,
                "could not find the vendor-data configuration file, continuing without it",
            )
        )
        vendor_data_path = None

    if not exists(network_config_path):
        verbose_outputs.append(
            PATH_NOT_FOUND_ERROR_MSG.format(
                network_config_path,
                "could not find the network-config configuration file, continuing without it",
            )
        )
        network_config_path = None
    if verbose:
        verbose_outputs.append(
            "Generating the cloud-init iso image at: {}".format(
                cloud_init_iso_output_path
            )
        )

    generated_result, generated_msg = generate_image_configuration(
        cloud_init_iso_output_path,
        user_data_path=user_data_path,
        meta_data_path=meta_data_path,
        vendor_data_path=vendor_data_path,
        network_config_path=network_config_path,
    )
    if verbose:
        verbose_outputs.append(generated_msg)
    if not generated_result:
        response["msg"] = generated_msg
        response["verbose_outputs"] = verbose_outputs
        return generated_result, response

    if not exists(os.path.dirname(configure_vm_log_path)):
        created = makedirs(os.path.dirname(configure_vm_log_path))
        if not created:
            response["msg"] = PATH_CREATE_ERROR_MSG.format(
                os.path.dirname(configure_vm_log_path)
            )
            response["verbose_outputs"] = verbose_outputs
            return PATH_CREATE_ERROR, response

    incrementer = 0
    while exists(configure_vm_log_path):
        if incrementer == 0:
            if verbose:
                verbose_outputs.append(
                    f"The configuring log file: {configure_vm_log_path} already exists, increasing the designated file name"
                )
            configure_vm_log_path = f"{configure_vm_log_path}.%s" % incrementer
        else:
            file_increment = os.path.splitext(configure_vm_log_path)[1]
            configure_vm_log_path = configure_vm_log_path.replace(
                file_increment, f".{incrementer + 1}"
            )
        incrementer += 1
    if verbose:
        verbose_outputs.append(f"Generated new log file path: {configure_vm_log_path}")
        verbose_outputs.append(
            f"Using the VM template description: {configure_vm_template_path}"
        )

    # TODO, these does not validate the template values correctly
    # Ensure that the required template values are set for the cloud-init iso image
    # and for the VM log file that is monitored to tell when the configuration process is finished
    if "num_vcpus" not in configure_vm_template_values:
        configure_vm_template_values["num_vcpus"] = CONFIGURE_VM_VCPUS
    if "memory_size" not in configure_vm_template_values:
        configure_vm_template_values["memory_size"] = CONFIGURE_VM_MEMORY
    if "cpu_architecture" not in configure_vm_template_values:
        configure_vm_template_values["cpu_architecture"] = CPU_ARCHITECTURE
    if "machine" not in configure_vm_template_values:
        configure_vm_template_values["machine"] = CONFIGURE_VM_MACHINE
    if "cd_iso_path" not in configure_vm_template_values:
        configure_vm_template_values["cd_iso_path"] = cloud_init_iso_output_path
    if (
        "configure_vm_log_path" not in configure_vm_template_values
        and "log_file_path" not in configure_vm_template_values
    ):
        configure_vm_template_values["log_file_path"] = configure_vm_log_path

    configured_id, configured_msg = await configure_image(
        configure_vm_name,
        image_path,
        template_path=configure_vm_template_path,
        **configure_vm_template_values,
    )
    if verbose:
        verbose_outputs.append(configured_msg)
    if not configured_id:
        response["msg"] = CONFIGURE_IMAGE_ERROR_MSG.format(
            image_path, "failed to configure image"
        )
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response

    if verbose:
        verbose_outputs.append("Waiting for the configuration process to finish")
    finished = finished_configure(configure_vm_log_path)
    if not finished:
        response["msg"] = "Failed to finish configuring the image"
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response
    if verbose:
        verbose_outputs.append(
            f"Finished configuring the image in the instance: {configured_id}"
        )

    shutdown, shutdown_msg = vm_action("stop", configured_id)
    if not shutdown:
        response["msg"] = (
            f"Failed to shutdown the VM: {configured_id} after configuration: {shutdown_msg}"
        )
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response

    shutdowned, shutdowned_msg = wait_for_vm_shutdown(configured_id)
    if not shutdowned:
        response["msg"] = (
            f"Failed to wait for the shutdown of VM: {configured_id} after configuration: {shutdowned_msg}"
        )
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response

    remove, remove_msg = vm_action("remove", configured_id)
    if not remove:
        response["msg"] = (
            f"Failed to remove the VM: {configured_id} after configuration: {remove_msg}"
        )
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response

    removed, removed_msg = wait_for_vm_removed(configured_id)
    if not removed:
        response["msg"] = (
            f"Failed to wait for the removal of VM: {configured_id} after the configuration was applied: {removed_msg}"
        )
        response["verbose_outputs"] = verbose_outputs
        return CONFIGURE_IMAGE_ERROR, response
    if verbose:
        verbose_outputs.append(
            f"Removed the VM: {configured_id} after configuration: {removed_msg}"
        )

    reset_success, reset_results = reset_image(
        image_path, reset_operations=reset_operations, verbose=verbose
    )
    if verbose:
        verbose_outputs.append(reset_results)
    if not reset_success:
        response["msg"] = RESET_IMAGE_ERROR_MSG.format(
            reset_results, "failed to reset image"
        )
        response["verbose_outputs"] = verbose_outputs
        return RESET_IMAGE_ERROR, response
    response["msg"] = "Succesfully configured image: {}".format(image_path)
    return SUCCESS, response



import os
import shutil
import re


def makedirs(path):
    try:
        os.makedirs(os.path.expanduser(path))
        return True
    except Exception:
        # TODO, add logging
        return False
    return False


def load(path, mode="r", readlines=False, handler=None, **load_kwargs):
    try:
        with open(path, mode) as fh:
            if handler:
                return handler.load(fh, **load_kwargs)
            if readlines:
                return fh.readlines()
            return fh.read()
    except Exception:
        # TODO, add logging
        return False
    return False


def write(path, content, mode="w", mkdirs=False, handler=None, **handler_kwargs):
    dir_path = os.path.dirname(path)
    if not os.path.exists(dir_path) and mkdirs:
        if not makedirs(dir_path):
            return False
    try:
        with open(path, mode) as fh:
            if handler:
                handler.dump(content, fh, **handler_kwargs)
            else:
                if isinstance(content, (list, set)):
                    for line in content:
                        fh.write(line)
                else:
                    fh.write(content)
        return True
    except Exception:
        # TODO, add logging
        return False
    return False


def remove(path, recursive=False):
    try:
        if recursive:
            shutil.rmtree(path)
        else:
            os.remove(path)
        return True
    except Exception:
        # TODO, add logging
        return False
    return False


def exists(path):
    return os.path.exists(os.path.expanduser(path))


def join(path, *paths):
    return os.path.join(path, *paths)


def which(command):
    return shutil.which(command)


def chmod(path, mode, **kwargs):
    try:
        os.chmod(os.path.expanduser(path), mode, **kwargs)
        return True
    except Exception:
        # TODO, add logging
        return False
    return False


def chown(path, uid, gid):
    try:
        os.chown(os.path.expanduser(path), uid, gid)
        return True
    except Exception:
        # TODO, add logging
        return False
    return False


def copy(original, target):
    # Copy path to target
    try:
        shutil.copyfile(original, target)
        return True
    except Exception:
        return False
    return False


# Read chunks of a file, default to 64KB
def hashsum(path, algorithm="sha1", buffer_size=65536):
    try:
        import hashlib

        hash_algorithm = hashlib.new(algorithm)
        with open(path, "rb") as fh:
            for chunk in iter(lambda: fh.read(buffer_size), b""):
                hash_algorithm.update(chunk)
        return hash_algorithm.hexdigest()
    except Exception:
        # TODO, add logging
        return False
    return False


def find(directory_path, regex_name):
    found = []
    for root, dirs, files in os.walk(directory_path):
        for f in files:
            if re.match(regex_name, f):
                found.append(f)
    return found


def size(path):
    try:
        return os.path.getsize(path)
    except Exception:
        # TODO, add logging
        return False
    return False



import json
import subprocess
import datetime


def __to_str__(o):
    if hasattr(o, "asdict"):
        return o.asdict()
    if isinstance(o, datetime.datetime):
        return o.__str__()
    if isinstance(o, bytes):
        return o.decode("utf-8")


def __format_output__(output, to_format="str"):
    if to_format == "str":
        if isinstance(output, bytes):
            return output.decode("utf-8")
        if isinstance(output, list):
            return ",".join(output)
        if isinstance(output, dict):
            formatted_output = {}
            for key, value in output.items():
                formatted_output[key] = str(value)
            return json.dumps(formatted_output)
        if isinstance(output, str):
            return output
    if to_format == "json":
        if is_json_string(output):
            return to_json(output)
        else:
            raise ValueError(f"Failed to convert output to json: {output}")
    return False


def __extract_results__(result):
    command_results = {"output": "", "error": "", "returncode": 1}
    if hasattr(result, "args"):
        command_results.update({"command": " ".join((getattr(result, "args")))})
    if hasattr(result, "returncode"):
        command_results.update({"returncode": getattr(result, "returncode")})
    if hasattr(result, "stderr"):
        command_results.update({"error": getattr(result, "stderr")})
    if hasattr(result, "stdout"):
        command_results.update({"output": getattr(result, "stdout")})
    if hasattr(result, "wait"):
        command_results.update({"wait": getattr(result, "wait")})
    if hasattr(result, "communicate"):
        command_results.update({"communicate": getattr(result, "communicate")})
    if hasattr(result, "kill"):
        command_results.update({"kill": getattr(result, "kill")})
    return command_results


def is_json_string(content):
    try:
        json.loads(content)
        return True
    except json.JSONDecodeError:
        return False
    return False


def to_json(content):
    return json.loads(content)


def run_popen(cmd, output_format="str", **run_kwargs):
    result = __extract_results__(subprocess.Popen(cmd, **run_kwargs))
    return __format_output__(result, to_format=output_format)


def check_call(cmd, output_format="str", **run_kwargs):
    result = __extract_results__(subprocess.check_call(cmd, **run_kwargs))
    return __format_output__(result, to_format=output_format)


def run(cmd, output_format="str", **run_kwargs):
    if not output_format:
        output_format = "str"
    return_values = {"output": "", "error": ""}
    try:
        raw_results = subprocess.run(cmd, **run_kwargs, capture_output=True)
    except Exception as e:
        return_values["error"] = f"Failed to run command: {cmd}, error: {e}"
        return False, return_values

    result = __extract_results__(raw_results)
    if result["error"]:
        formatted_error = __format_output__(result["error"], to_format=output_format)
        if formatted_error:
            return_values["error"] = formatted_error

    if result["output"]:
        formatted_output = __format_output__(result["output"], to_format=output_format)
        if formatted_output:
            return_values["output"] = formatted_output

    if result["returncode"] != 0:
        return False, return_values
    return True, return_values



import os
from configure_vm_image.common.defaults import (
    CLOUD_INIT_DIR,
    TMP_DIR,
    RES_DIR,
    CONFIGURE_ARGUMENT,
)
from configure_vm_image.cli.parsers.actions import (
    PositionalArgumentsAction,
    KeyValueAction,
)


def configure_group(parser):
    configure_group_ = parser.add_argument_group(
        title="Configure Virtual Machine Image"
    )
    configure_group_.add_argument(
        "image_path",
        action=PositionalArgumentsAction,
        help="The path to the image that is to be configured.",
    )
    configure_group_.add_argument(
        "--image-format",
        dest="{}_image_format".format(CONFIGURE_ARGUMENT),
        default=None,
        help="The format of the image that is to be configured. The tool tries to automatically discover this if not set.",
    )
    configure_group_.add_argument(
        "--config-user-data-path",
        dest="{}_user_data_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(CLOUD_INIT_DIR, "user-data"),
        help="The path to the cloud-init user-data configuration file.",
    )
    configure_group_.add_argument(
        "--config-meta-data-path",
        dest="{}_meta_data_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(CLOUD_INIT_DIR, "meta-data"),
        help="The path to the cloud-init meta-data configuration file.",
    )
    configure_group_.add_argument(
        "--config-vendor-data-path",
        dest="{}_vendor_data_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(CLOUD_INIT_DIR, "vendor-data"),
        help="The path to the cloud-init vendor-data configuration file.",
    )
    configure_group_.add_argument(
        "--config-network-config-path",
        dest="{}_network_config_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(CLOUD_INIT_DIR, "network-config"),
        help="""The path to the cloud-init network-config configuration file
        that is used to configure the network settings of the image.
        """,
    )
    configure_group_.add_argument(
        "--configure-vm-orchestrator",
        "-cv-orch",
        dest="{}_configure_vm_orchestrator".format(CONFIGURE_ARGUMENT),
        default="libvirt-provider",
        help="The orchestrator to use when provisioning the virtual machine that is used to configure a particular virtual machine image",
    )
    configure_group_.add_argument(
        "--configure-vm-name",
        "-cv-name",
        dest="{}_configure_vm_name".format(CONFIGURE_ARGUMENT),
        default="configure-vm-image",
        help="The name of the VM that is used to configure the image.",
    )
    configure_group_.add_argument(
        "--cloud-init-iso-output-path",
        "-ci-output",
        dest="{}_cloud_init_iso_output_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(CLOUD_INIT_DIR, "cidata.iso"),
        help="""The path to the cloud-init output iso image file that is generated
        based on the data defined in the user-data, meta-data, vendor-data, and network-config files.
        This seed iso file is then subsequently used to configure the defined input image.
        """,
    )
    configure_group_.add_argument(
        "--configure-vm-log-path",
        "-cv-log",
        dest="{}_configure_vm_log_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(TMP_DIR, "configure-vm.log"),
        help="""The path to the log file that is used to log the output of the configuring VM.""",
    )
    configure_group_.add_argument(
        "--configure-vm-template-path",
        "-cv-tp",
        dest="{}_configure_vm_template_path".format(CONFIGURE_ARGUMENT),
        default=os.path.join(RES_DIR, "configure-vm-template.xml.j2"),
        help="""The path to the template file that is used to configure the VM.""",
    )
    configure_group_.add_argument(
        "--configure-vm-template-values",
        "-cv-tv",
        dest="{}_configure_vm_template_values".format(CONFIGURE_ARGUMENT),
        metavar="KEY=VALUE",
        action=KeyValueAction,
        default="",
        help="""An additional set of comma seperated KEY=VALUE pair arguments that should be passed to the --configure-vm-template-path.
        If a value contains spaces, you should define it with quotes.
        """,
    )
    configure_group_.add_argument(
        "--reset-operations",
        "-ro",
        dest="{}_reset_operations".format(CONFIGURE_ARGUMENT),
        default="defaults,-ssh-userdir",
        help="""The operations to perform during the reset operation.""",
    )
    configure_group_.add_argument(
        "--verbose",
        "-v",
        dest="{}_verbose".format(CONFIGURE_ARGUMENT),
        action="store_true",
        default=False,
        help="Flag to enable verbose output.",
    )



SUCCESS = 0
PATH_NOT_FOUND_ERROR = 1
PATH_NOT_FOUND_ERROR_MSG = "Path not found: {} - error: {}"
PATH_LOAD_ERROR = 2
PATH_LOAD_ERROR_MSG = "Failed to load path: {}"
PATH_CREATE_ERROR = 3
PATH_CREATE_ERROR_MSG = "Failed to create path: {}"
CONFIGURE_IMAGE_ERROR = 4
CONFIGURE_IMAGE_ERROR_MSG = "Failed to configure image: {} - error: {}"
RESET_IMAGE_ERROR = 5
RESET_IMAGE_ERROR_MSG = "Failed to reset image: {} - error: {}"

MISSING_ATTRIBUTE_ERROR = 6
MISSING_ATTRIBUTE_ERROR_MSG = "Missing attribute: {} in {}"
INVALID_ATTRIBUTE_TYPE_ERROR = 7
INVALID_ATTRIBUTE_TYPE_ERROR_MSG = (
    "Invalid attribute type: {} for value: {} - must be {}"
)
CHECKSUM_ERROR = 8
RESIZE_ERROR = 9
RESIZE_ERROR_MSG = "Failed to resize path: {}"
CHECK_ERROR = 10
CHECK_ERROR_MSG = "Failed to check path: {}"
JSON_DUMP_ERROR = 11
JSON_DUMP_ERROR_MSG = "Failed to dump JSON: {}"
DOWNLOAD_ERROR = 12
FUNCTION_NOT_FOUND_ERROR = 13



def strip_argument_prefix(arguments, prefix=""):
    return {k.replace(prefix, ""): v for k, v in arguments.items()}


def get_arguments(arguments, startswith=""):
    return {k: v for k, v in arguments.items() if k.startswith(startswith)}


def extract_arguments(arguments, argument_groups):
    found_kwargs, remaining_kwargs = {}, {}
    for argument_group in argument_groups:
        group_args = get_arguments(arguments, argument_group.lower())
        found_kwargs.update(group_args)
    remaining_kwargs = {
        k: v for k, v in arguments.items() if k not in found_kwargs and v
    }
    return found_kwargs, remaining_kwargs


def strip_argument_group_prefix(arguments, argument_groups):
    args = {}
    for argument_group in argument_groups:
        group_arguments = get_arguments(arguments, argument_group.lower())
        args.update(
            strip_argument_prefix(group_arguments, argument_group.lower() + "_")
        )
    return args



import datetime
import os
import sys


def error_print(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def to_str(o):
    if hasattr(o, "asdict"):
        return o.asdict()
    if isinstance(o, datetime.datetime):
        return o.__str__()
    if isinstance(o, bytes):
        return o.decode("utf-8")
    return o


def expand_path(path):
    return os.path.realpath(os.path.expanduser(path))



import unittest
import os
import random
from configure_vm_image.utils.io import exists, join, copy, remove
from configure_vm_image.cli.configure_image import main
from configure_vm_image.common.defaults import (
    CPU_ARCHITECTURE,
    CONFIGURE_VM_MACHINE,
    CONFIGURE_VM_MEMORY,
    CONFIGURE_VM_VCPUS,
)
from configure_vm_image.common.codes import SUCCESS
from .context import AsyncConfigureTestContext


def cli_action(*args):
    return_code = None
    try:
        return_code = main(args)
    except SystemExit as e:
        return_code = e.code
    return return_code


class TestCLIConfigurer(unittest.IsolatedAsyncioTestCase):
    context = AsyncConfigureTestContext()

    async def asyncSetUp(self):
        await self.context.setUp()

        self.seed = str(random.random())[2:10]
        self.assertTrue(exists(self.context.image))

        self.configure_vm_name = f"configure-vm-test-{self.seed}"
        self.image_to_configure = os.path.realpath(
            f"{os.path.splitext(self.context.image)[0]}-{self.seed}.{self.context.image_config['output_format']}"
        )
        self.cloud_init_directory = join(
            self.context.test_res_directory, "test-cloud-init"
        )
        self.cloud_init_output_directory = join(
            self.context.test_tmp_directory, "cloud-init"
        )
        self.assertTrue(copy(self.context.image, self.image_to_configure))
        self.assertTrue(exists(self.image_to_configure))

        self.configure_vm_log_path = os.path.realpath(
            f"{join(self.context.test_tmp_directory, self.configure_vm_name)}.log"
        )

    async def asyncTearDown(self):
        # Cleanup the test images
        self.assertTrue(remove(self.image_to_configure))

    @classmethod
    def tearDownClass(cls):
        cls.context.tearDown()

    def test_configurer_cli_with_default_template(self):
        args = [
            self.image_to_configure,
            "--config-user-data-path",
            join(self.cloud_init_directory, "user-data"),
            "--config-network-config-path",
            join(self.cloud_init_directory, "network-config"),
            "--cloud-init-iso-output-path",
            join(self.cloud_init_output_directory, f"{self.seed}-cidata.iso"),
            "--configure-vm-name",
            self.configure_vm_name,
            "--configure-vm-log-path",
            self.configure_vm_log_path,
            "--configure-vm-template-path",
            self.context.image_template_config,
        ]
        return_code = cli_action(*args)
        self.assertEqual(return_code, SUCCESS)

    def test_configurer_cli_with_set_template_values(self):
        configure_vm_template_values = ",".join(
            [
                f"memory_size={CONFIGURE_VM_MEMORY}",
                f"num_vcpus={CONFIGURE_VM_VCPUS}",
                f"cpu_architecture={CPU_ARCHITECTURE}",
                f"machine={CONFIGURE_VM_MACHINE}",
            ]
        )

        args = [
            self.image_to_configure,
            "--config-user-data-path",
            join(self.cloud_init_directory, "user-data"),
            "--config-network-config-path",
            join(self.cloud_init_directory, "network-config"),
            "--cloud-init-iso-output-path",
            join(self.cloud_init_output_directory, f"{self.seed}-cidata.iso"),
            "--configure-vm-name",
            self.configure_vm_name,
            "--configure-vm-log-path",
            self.configure_vm_log_path,
            "--configure-vm-template-values",
            configure_vm_template_values,
            "--configure-vm-template-path",
            self.context.image_template_config,
        ]
        return_code = cli_action(*args)
        self.assertEqual(return_code, SUCCESS)



import unittest
import os
import random
from configure_vm_image.configure import configure_vm_image
from configure_vm_image.utils.io import join, copy, exists, remove
from configure_vm_image.common.defaults import (
    CPU_ARCHITECTURE,
    CONFIGURE_VM_MACHINE,
    CONFIGURE_VM_MEMORY,
    CONFIGURE_VM_VCPUS,
)
from configure_vm_image.common.codes import SUCCESS
from .context import AsyncConfigureTestContext


class AsyncTestImageConfiguration(unittest.IsolatedAsyncioTestCase):
    context = AsyncConfigureTestContext()

    async def asyncSetUp(self):
        await self.context.setUp()

        self.seed = str(random.random())[2:10]
        self.assertTrue(exists(self.context.image))

        self.configure_vm_name = f"configure-vm-test-{self.seed}"
        self.image_to_configure = os.path.realpath(
            f"{os.path.splitext(self.context.image)[0]}-{self.seed}.{self.context.image_config['output_format']}"
        )
        self.cloud_init_directory = join(
            self.context.test_res_directory, "test-cloud-init"
        )
        self.cloud_init_output_directory = join(
            self.context.test_tmp_directory, "cloud-init"
        )

        self.assertTrue(copy(self.context.image, self.image_to_configure))
        self.assertTrue(exists(self.image_to_configure))

        self.configure_vm_log_path = os.path.realpath(
            f"{join(self.context.test_tmp_directory, self.configure_vm_name)}.log"
        )

    async def asyncTearDown(self):
        # Cleanup the test images
        self.assertTrue(remove(self.image_to_configure))

    @classmethod
    def tearDownClass(cls):
        cls.context.tearDown()

    async def test_basic_configure(self):
        return_code, msg = await configure_vm_image(
            self.image_to_configure,
            network_config_path=join(self.cloud_init_directory, "network-config"),
            user_data_path=join(self.cloud_init_directory, "user-data"),
            cloud_init_iso_output_path=join(
                self.cloud_init_output_directory, f"{self.seed}-cidata.iso"
            ),
            configure_vm_name=self.configure_vm_name,
            configure_vm_log_path=self.configure_vm_log_path,
            configure_vm_template_path=self.context.image_template_config,
            verbose=True,
        )
        self.assertEqual(return_code, SUCCESS)
        self.assertIsNotNone(msg)

    async def test_basic_configure_with_template_values(self):
        configure_vm_template_values = {
            "memory_size": CONFIGURE_VM_MEMORY,
            "num_vcpus": CONFIGURE_VM_VCPUS,
            "cpu_architecture": CPU_ARCHITECTURE,
            "machine": CONFIGURE_VM_MACHINE,
        }

        return_code, msg = await configure_vm_image(
            self.image_to_configure,
            network_config_path=join(self.cloud_init_directory, "network-config"),
            user_data_path=join(self.cloud_init_directory, "user-data"),
            cloud_init_iso_output_path=join(
                self.cloud_init_output_directory, f"{self.seed}-cidata.iso"
            ),
            configure_vm_name=self.configure_vm_name,
            configure_vm_log_path=self.configure_vm_log_path,
            configure_vm_template_path=self.context.image_template_config,
            configure_vm_template_values=configure_vm_template_values,
            verbose=True,
        )
        self.assertEqual(return_code, SUCCESS)
        self.assertIsNotNone(msg)
